#!/usr/bin/env python3
import argparse
import sys
import os
from datetime import datetime
import subprocess

from adaptive_hybrid_sort.enhanced_adaptive_sort import EnhancedAdaptiveSort
from adaptive_hybrid_sort.sorting_visualizer import SortingVisualizer

def setup_environment():
    """Setup the required environment and install dependencies."""
    # Create results directory structure
    base_dir = "sorting_results"
    date_str = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    result_dir = os.path.join(base_dir, date_str)
    
    os.makedirs(result_dir, exist_ok=True)
    print(f"Created results directory: {result_dir}")
    
    # Install requirements
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "install", "-r", "requirements.txt"])
        print("Successfully installed requirements")
    except subprocess.CalledProcessError:
        print("Warning: Failed to install some requirements")
    
    return result_dir

def run_benchmarks(args, result_dir):
    """Run sorting benchmarks with specified parameters."""
    visualizer = SortingVisualizer()
    
    # Update file paths to use result directory
    visualizer.result_dir = result_dir
    
    print(f"\nRunning benchmarks with:")
    print(f"Sizes: {args.sizes}")
    print(f"Patterns: {args.patterns}")
    print(f"Runs per configuration: {args.runs}")
    
    visualizer.run_comprehensive_benchmarks(
        sizes=args.sizes,
        patterns=args.patterns,
        num_runs=args.runs
    )
    
    print(f"\nResults saved in: {result_dir}")
    print("Files generated:")
    print(f"- {os.path.join(result_dir, 'sorting_performance.png')}")
    print(f"- {os.path.join(result_dir, 'sorting_results.csv')}")
    print(f"- {os.path.join(result_dir, 'sorting_statistics.csv')}")

def sort_array(args):
    """Sort a single array with specified pattern."""
    sorter = EnhancedAdaptiveSort()
    visualizer = SortingVisualizer()
    
    # Generate test data
    arr = visualizer._generate_data(args.size, args.pattern).tolist()
    
    # Sort and measure performance
    print(f"\nSorting array with pattern '{args.pattern}' and size {args.size}")
    print("Sample of input:", arr[:10], "..." if len(arr) > 10 else "")
    
    start_time = datetime.now()
    sorted_arr = sorter.sort(arr)
    end_time = datetime.now()
    
    print("Sample of output:", sorted_arr[:10], "..." if len(sorted_arr) > 10 else "")
    print(f"Time taken: {(end_time - start_time).total_seconds():.4f} seconds")
    print("Sorting successful!")

def main():
    parser = argparse.ArgumentParser(
        description="Enhanced Adaptive Sort - A hybrid sorting algorithm with pattern detection"
    )
    
    subparsers = parser.add_subparsers(dest='command', help='Available commands')
    
    # Benchmark command
    bench_parser = subparsers.add_parser('benchmark', help='Run comprehensive benchmarks')
    bench_parser.add_argument(
        '--sizes',
        type=int,
        nargs='+',
        default=[1000, 2000, 4000, 8000, 16000],
        help='Array sizes to test'
    )
    bench_parser.add_argument(
        '--patterns',
        type=str,
        nargs='+',
        default=["random", "nearly_sorted", "reversed", "few_unique", "sorted", "pipeline"],
        help='Patterns to test'
    )
    bench_parser.add_argument(
        '--runs',
        type=int,
        default=5,
        help='Number of runs per configuration'
    )
    
    # Sort command
    sort_parser = subparsers.add_parser('sort', help='Sort a single array')
    sort_parser.add_argument(
        '--size',
        type=int,
        default=1000,
        help='Size of the array to sort'
    )
    sort_parser.add_argument(
        '--pattern',
        type=str,
        default="random",
        choices=["random", "nearly_sorted", "reversed", "few_unique", "sorted", "pipeline"],
        help='Pattern of the array to sort'
    )
    
    # Setup command
    subparsers.add_parser('setup', help='Install requirements and setup environment')
    
    args = parser.parse_args()
    
    if args.command == 'setup':
        setup_environment()
    elif args.command == 'benchmark':
        result_dir = setup_environment()
        run_benchmarks(args, result_dir)
    elif args.command == 'sort':
        sort_array(args)
    else:
        parser.print_help()

if __name__ == "__main__":
    main() 