import random
import time
from typing import List, TypeVar, Optional, Union
import statistics
import math
import numpy as np
from concurrent.futures import ThreadPoolExecutor
import threading
from contextlib import contextmanager
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from memory_profiler import memory_usage
from tqdm import tqdm
import os

T = TypeVar('T', int, float)

class EnhancedAdaptiveSort:
    def __init__(self, threshold_small: int = 32, threshold_chunk: int = 2048, 
                 parallel_threshold: int = 8192):
        self.threshold_small = threshold_small
        self.threshold_chunk = threshold_chunk
        self.sample_size = min(30, int(math.log2(threshold_chunk)))
        self.parallel_threshold = parallel_threshold
        self.num_threads = max(1, min(32, threading.active_count() * 2))
        
    def _insertion_sort(self, arr: np.ndarray, start: int, end: int) -> None:
        """Optimized insertion sort."""
        if end - start <= 1:
            return
            
        # Use NumPy's searchsorted for better performance
        for i in range(start + 1, end):
            key = arr[i]
            insert_pos = np.searchsorted(arr[start:i], key) + start
            if insert_pos < i:
                arr[insert_pos+1:i+1] = arr[insert_pos:i]
                arr[insert_pos] = key

    def _partition_three_way(self, arr: np.ndarray, low: int, high: int) -> tuple[int, int]:
        """Three-way partitioning optimized for NumPy."""
        if high - low <= 1:
            return low, high
            
        # Use median-of-three for better pivot selection
        mid = (low + high) // 2
        if arr[low] > arr[mid]: arr[low], arr[mid] = arr[mid], arr[low]
        if arr[low] > arr[high]: arr[low], arr[high] = arr[high], arr[low]
        if arr[mid] > arr[high]: arr[mid], arr[high] = arr[high], arr[mid]
        pivot = arr[mid]
        
        # Use NumPy operations for partitioning
        mask_less = arr[low:high+1] < pivot
        mask_equal = arr[low:high+1] == pivot
        mask_greater = arr[low:high+1] > pivot
        
        # Count elements in each partition
        count_less = np.sum(mask_less)
        count_equal = np.sum(mask_equal)
        
        # Create partitioned array
        temp = np.empty_like(arr[low:high+1])
        temp[:count_less] = arr[low:high+1][mask_less]
        temp[count_less:count_less+count_equal] = arr[low:high+1][mask_equal]
        temp[count_less+count_equal:] = arr[low:high+1][mask_greater]
        
        # Copy back to original array
        arr[low:high+1] = temp
        
        return low + count_less, low + count_less + count_equal - 1

    def _quick_sort(self, arr: np.ndarray, low: int, high: int) -> None:
        """Optimized quicksort implementation."""
        while low < high:
            if high - low + 1 <= self.threshold_small:
                self._insertion_sort(arr, low, high + 1)
                break
                
            lt, gt = self._partition_three_way(arr, low, high)
            
            # Handle the smaller partition first
            if lt - low < high - gt:
                self._quick_sort(arr, low, lt - 1)
                low = gt + 1
            else:
                self._quick_sort(arr, gt + 1, high)
                high = lt - 1

    def _merge(self, left: np.ndarray, right: np.ndarray) -> np.ndarray:
        """Optimized merge operation using NumPy."""
        if len(left) == 0: return right
        if len(right) == 0: return left
        
        # For small arrays, use NumPy's sort
        if len(left) + len(right) <= self.threshold_chunk:
            result = np.concatenate([left, right])
            result.sort()
            return result
        
        # Allocate result array
        total_len = len(left) + len(right)
        result = np.empty(total_len, dtype=left.dtype)
        
        # Use NumPy's searchsorted for efficient merging
        if len(left) < len(right):
            # Insert left elements into right
            indices = np.searchsorted(right, left)
            # Create a mask of valid positions
            mask = np.ones(total_len, dtype=bool)
            valid_indices = indices + np.arange(len(indices))
            valid_indices = valid_indices[valid_indices < total_len]
            mask[valid_indices] = False
            # Place elements
            result[valid_indices] = left[:len(valid_indices)]
            result[mask] = right
        else:
            # Insert right elements into left
            indices = np.searchsorted(left, right)
            result[:len(left)] = left
            # Shift elements to make space
            for i, insert_idx in enumerate(indices):
                if insert_idx < len(result):
                    result[insert_idx+1:] = result[insert_idx:-1]
                    result[insert_idx] = right[i]
        
        return result

    def _sort_chunk(self, chunk: np.ndarray) -> np.ndarray:
        """Sort a single chunk with optimal algorithm selection."""
        chunk_copy = chunk.copy()  # Create a copy to avoid modifying the original
        if len(chunk_copy) <= self.threshold_small:
            self._insertion_sort(chunk_copy, 0, len(chunk_copy))
        else:
            # Use quicksort for better stability
            self._quick_sort(chunk_copy, 0, len(chunk_copy) - 1)
        return chunk_copy

    def _parallel_sort(self, arr: np.ndarray) -> np.ndarray:
        """Parallel sorting with improved chunk handling."""
        if len(arr) <= self.parallel_threshold:
            return self._sort_chunk(arr)

        # Calculate optimal chunk size based on CPU cache size
        chunk_size = max(
            self.parallel_threshold,
            min(len(arr) // (self.num_threads * 2),
                len(arr) // 4)
        )
        
        # Split into chunks
        chunks = [
            arr[i:i + chunk_size].copy()
            for i in range(0, len(arr), chunk_size)
        ]
        
        # Sort chunks in parallel
        with ThreadPoolExecutor(max_workers=self.num_threads) as executor:
            # Sort each chunk
            sorted_chunks = []
            for chunk in chunks:
                sorted_chunk = executor.submit(self._sort_chunk, chunk)
                sorted_chunks.append(sorted_chunk)
            
            # Wait for all chunks to complete
            sorted_chunks = [chunk.result() for chunk in sorted_chunks]
            
            # Merge pairs of chunks in parallel
            while len(sorted_chunks) > 1:
                new_chunks = []
                futures = []
                
                for i in range(0, len(sorted_chunks), 2):
                    if i + 1 < len(sorted_chunks):
                        future = executor.submit(
                            self._merge, 
                            sorted_chunks[i], 
                            sorted_chunks[i + 1]
                        )
                        futures.append(future)
                    else:
                        new_chunks.append(sorted_chunks[i])
                
                # Wait for merges to complete
                sorted_chunks = [f.result() for f in futures] + new_chunks
        
        return sorted_chunks[0]

    def _detect_pattern(self, arr: np.ndarray) -> tuple[float, bool, bool]:
        """Pattern detection using NumPy operations."""
        if len(arr) < self.threshold_small:
            return 0.0, False, False
            
        # Quick checks for common patterns
        if len(arr) > 1000:
            # Use strided sampling for efficiency
            stride = max(1, len(arr) // 100)
            samples = arr[::stride]
            diffs = np.diff(samples)
            
            # Vectorized pattern detection
            is_sorted = np.all(diffs >= 0)
            is_reversed = np.all(diffs <= 0)
            
            if is_sorted:
                return 0.0, True, False
            if is_reversed:
                return 0.0, False, True
            
            # Check for nearly sorted
            disorder_ratio = np.sum(diffs < 0) / len(diffs)
            if disorder_ratio < 0.1:
                return disorder_ratio, True, False
        
        return 1.0, False, False

    def sort(self, arr: List[T]) -> List[T]:
        """Main sorting interface with improved type handling."""
        if not arr:
            return arr
            
        try:
            # Determine input type and numpy dtype
            input_type = type(arr[0])
            if input_type in (int, np.int32, np.int64):
                dtype = np.int64
            elif input_type in (float, np.float32, np.float64):
                dtype = np.float64
            else:
                raise ValueError(f"Unsupported type: {input_type}")
            
            # Convert to numpy array efficiently
            np_arr = np.asarray(arr, dtype=dtype)
            
            # Detect patterns and choose strategy
            disorder_ratio, is_nearly_sorted, is_reversed = self._detect_pattern(np_arr)
            
            if is_reversed:
                np_arr = np.flip(np_arr)
            elif is_nearly_sorted:
                self._insertion_sort(np_arr, 0, len(np_arr))
            elif len(np_arr) <= self.threshold_small:
                self._insertion_sort(np_arr, 0, len(np_arr))
            else:
                np_arr = self._parallel_sort(np_arr)
            
            # Convert back preserving type
            return [input_type(x) for x in np_arr]
            
        except Exception as e:
            print(f"Error during sorting: {str(e)}")
            # Fallback to Python's built-in sort
            arr_copy = arr.copy()
            arr_copy.sort()
            return arr_copy


class SortingVisualizer:
    def __init__(self):
        self.sorter = EnhancedAdaptiveSort()
        plt.style.use('default')
        sns.set_theme(style="whitegrid")
        self.result_dir = "."  # Default to current directory
        
    def _generate_data(self, size: int, pattern: str) -> np.ndarray:
        """Generate test data with various patterns."""
        if pattern == "random":
            return np.random.randint(0, 1000000, size=size)
        elif pattern == "nearly_sorted":
            arr = np.arange(size)
            perturbations = size // 20
            for _ in range(perturbations):
                i, j = np.random.randint(0, size, 2)
                arr[i], arr[j] = arr[j], arr[i]
            return arr
        elif pattern == "reversed":
            return np.arange(size-1, -1, -1)
        elif pattern == "few_unique":
            return np.random.choice(np.arange(10), size=size)
        elif pattern == "sorted":
            return np.arange(size)
        elif pattern == "pipeline":
            return np.concatenate([
                np.random.randint(0, 1000000, size=size//3),
                np.sort(np.random.randint(0, 1000000, size=size//3)),
                np.sort(np.random.randint(0, 1000000, size=size//3))[::-1]
            ])
        else:
            raise ValueError(f"Unknown pattern: {pattern}")

    def _measure_performance(self, arr: np.ndarray) -> dict:
        """Measure sorting performance metrics."""
        start_time = time.time()
        mem_usage = max(memory_usage((self.sorter.sort, (arr.tolist(),))))
        end_time = time.time()
        
        # Compare with numpy's sort
        np_arr = arr.copy()
        np_start = time.time()
        np.sort(np_arr)
        np_time = time.time() - np_start
        
        return {
            'time': end_time - start_time,
            'memory': mem_usage,
            'relative_speed': np_time / (end_time - start_time)
        }

    def run_comprehensive_benchmarks(self, sizes: List[int], patterns: List[str], num_runs: int = 5):
        """Run comprehensive benchmarks and generate visualizations."""
        results = []
        
        for size in tqdm(sizes, desc="Testing sizes"):
            for pattern in patterns:
                for run in range(num_runs):
                    arr = self._generate_data(size, pattern)
                    metrics = self._measure_performance(arr)
                    results.append({
                        'size': size,
                        'pattern': pattern,
                        'run': run,
                        'time': metrics['time'],
                        'memory': metrics['memory'],
                        'relative_speed': metrics['relative_speed']
                    })
        
        # Convert to DataFrame
        df = pd.DataFrame(results)
        
        # Save raw results
        df.to_csv(os.path.join(self.result_dir, 'sorting_results.csv'), index=False)
        
        # Calculate statistics
        stats = df.groupby(['size', 'pattern']).agg({
            'time': ['mean', 'std'],
            'memory': ['mean', 'std'],
            'relative_speed': ['mean', 'std']
        }).round(4)
        stats.to_csv(os.path.join(self.result_dir, 'sorting_statistics.csv'))
        
        # Create visualization
        plt.figure(figsize=(20, 15))
        plt.suptitle('Sorting Algorithm Performance Analysis', fontsize=16, y=0.95)
        
        # Time performance
        plt.subplot(2, 2, 1)
        sns.boxplot(data=df, x='size', y='time', hue='pattern')
        plt.title('Time Performance by Input Size and Pattern')
        plt.xticks(rotation=45)
        
        # Memory usage
        plt.subplot(2, 2, 2)
        sns.boxplot(data=df, x='size', y='memory', hue='pattern')
        plt.title('Memory Usage by Input Size and Pattern')
        plt.xticks(rotation=45)
        
        # Relative speed
        plt.subplot(2, 2, 3)
        sns.boxplot(data=df, x='pattern', y='relative_speed')
        plt.title('Relative Speed vs NumPy Sort')
        plt.xticks(rotation=45)
        
        # Time vs Size relationship
        plt.subplot(2, 2, 4)
        sns.lineplot(data=df, x='size', y='time', hue='pattern')
        plt.title('Time Complexity Analysis')
        plt.xticks(rotation=45)
        
        plt.tight_layout()
        plt.savefig(os.path.join(self.result_dir, 'sorting_performance.png'))
        plt.close()


# Testing suite
if __name__ == "__main__":
    def test_sorting(size: int, pattern: str = "random") -> None:
        if pattern == "random":
            arr = [random.randint(0, 1000000) for _ in range(size)]
        elif pattern == "nearly_sorted":
            arr = list(range(size))
            perturbations = size // 20
            for _ in range(perturbations):
                i = random.randint(0, size-1)
                j = random.randint(max(0, i-10), min(size-1, i+10))
                arr[i], arr[j] = arr[j], arr[i]
        elif pattern == "reversed":
            arr = list(range(size, 0, -1))
        elif pattern == "few_unique":
            arr = [random.randint(0, min(10, size//100)) for _ in range(size)]
        elif pattern == "sorted":
            arr = list(range(size))
        elif pattern == "pipeline":
            arr = [i if random.random() > 0.1 else random.randint(0, size) for i in range(size)]
        
        # Create copies for comparison
        arr_copy = arr.copy()
        arr_numpy = arr.copy()
        
        # Test our enhanced sorting
        sorter = EnhancedAdaptiveSort()
        start_time = time.time()
        sorted_arr = sorter.sort(arr)
        enhanced_time = time.time() - start_time
        
        # Test numpy's sort
        start_time = time.time()
        np.array(arr_numpy).sort()
        numpy_time = time.time() - start_time
        
        # Test Python's built-in sort
        start_time = time.time()
        arr_copy.sort()
        python_time = time.time() - start_time
        
        # Verify correctness
        assert sorted_arr == arr_copy, f"Sorting failed for pattern: {pattern}, size: {size}"
        
        print(f"\nPattern: {pattern}")
        print(f"Size: {size:,}")
        print(f"Enhanced Sort Time: {enhanced_time:.4f} seconds")
        print(f"NumPy Sort Time: {numpy_time:.4f} seconds")
        print(f"Python Sort Time: {python_time:.4f} seconds")
        print("-" * 60)
    
    # Run tests
    sizes = [1000, 10000, 100000, 1000000]
    patterns = ["random", "nearly_sorted", "reversed", "few_unique", "sorted", "pipeline"]
    
    print("\nRunning comprehensive sorting tests...")
    print("=" * 60)
    
    try:
        for size in sizes:
            for pattern in patterns:
                test_sorting(size, pattern)
    except KeyboardInterrupt:
        print("\nTesting interrupted by user")
    except Exception as e:
        print(f"\nError during testing: {str(e)}") 