import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from tqdm import tqdm
from .enhanced_adaptive_sort import EnhancedAdaptiveSort
import time
from typing import List, Dict, Any
from memory_profiler import profile
import psutil
import os
import random

class SortingVisualizer:
    def __init__(self):
        self.sorter = EnhancedAdaptiveSort()
        self.metrics: Dict[str, List[float]] = {
            'size': [],
            'pattern': [],
            'time': [],
            'memory': [],
            'comparisons': [],
            'swaps': [],
            'algorithm': []
        }
        # Set up plotting style
        plt.style.use('default')
        sns.set_theme(style="whitegrid")
        
    def _measure_memory(self, func, *args):
        """Measure peak memory usage of a function."""
        process = psutil.Process(os.getpid())
        mem_before = process.memory_info().rss
        result = func(*args)
        mem_after = process.memory_info().rss
        return result, (mem_after - mem_before) / 1024 / 1024  # Convert to MB

    def benchmark_sorting(self, sizes: List[int], patterns: List[str], 
                         repetitions: int = 3) -> pd.DataFrame:
        """Run comprehensive benchmarks with multiple metrics."""
        results = []
        total_iterations = len(sizes) * len(patterns) * repetitions
        
        with tqdm(total=total_iterations, desc="Running benchmarks") as pbar:
            for size in sizes:
                for pattern in patterns:
                    # Generate test data
                    if pattern == "random":
                        arr = [random.randint(0, 1000000) for _ in range(size)]
                    elif pattern == "nearly_sorted":
                        arr = list(range(size))
                        perturbations = size // 20
                        for _ in range(perturbations):
                            i = random.randint(0, size-1)
                            j = random.randint(max(0, i-10), min(size-1, i+10))
                            arr[i], arr[j] = arr[j], arr[i]
                    elif pattern == "reversed":
                        arr = list(range(size, 0, -1))
                    elif pattern == "few_unique":
                        arr = [random.randint(0, min(10, size//100)) for _ in range(size)]
                    elif pattern == "sorted":
                        arr = list(range(size))
                    elif pattern == "pipeline":
                        arr = [i if random.random() > 0.1 else random.randint(0, size) 
                              for i in range(size)]
                    
                    # Test different sorting algorithms
                    algorithms = {
                        'Enhanced': lambda x: self.sorter.sort(x.copy()),
                        'Python': lambda x: sorted(x.copy()),
                        'NumPy': lambda x: np.sort(np.array(x, dtype=np.int64))
                    }
                    
                    for algo_name, algo_func in algorithms.items():
                        times = []
                        memories = []
                        
                        for _ in range(repetitions):
                            # Measure time and memory
                            start_time = time.time()
                            _, mem_usage = self._measure_memory(algo_func, arr)
                            end_time = time.time()
                            
                            times.append(end_time - start_time)
                            memories.append(mem_usage)
                            pbar.update(1)
                        
                        # Record metrics
                        results.append({
                            'Size': size,
                            'Pattern': pattern,
                            'Algorithm': algo_name,
                            'Time': np.mean(times),
                            'Time_std': np.std(times),
                            'Memory': np.mean(memories),
                            'Memory_std': np.std(memories)
                        })
        
        return pd.DataFrame(results)

    def plot_performance_comparison(self, df: pd.DataFrame):
        """Create comprehensive performance visualization."""
        # Set up the figure
        fig = plt.figure(figsize=(20, 15))
        plt.suptitle('Sorting Algorithm Performance Analysis', size=16, y=0.95)
        
        # Set color palette
        colors = sns.color_palette("husl", n_colors=3)
        sns.set_palette(colors)
        
        # 1. Time comparison
        plt.subplot(3, 2, 1)
        g = sns.boxplot(data=df, x='Size', y='Time', hue='Algorithm')
        plt.yscale('log')
        plt.title('Sorting Time by Array Size')
        plt.xlabel('Array Size')
        plt.ylabel('Time (seconds)')
        g.legend(title='Algorithm', bbox_to_anchor=(1.05, 1), loc='upper left')
        
        # 2. Memory usage
        plt.subplot(3, 2, 2)
        g = sns.boxplot(data=df, x='Size', y='Memory', hue='Algorithm')
        plt.title('Memory Usage by Array Size')
        plt.xlabel('Array Size')
        plt.ylabel('Memory (MB)')
        g.legend(title='Algorithm', bbox_to_anchor=(1.05, 1), loc='upper left')
        
        # 3. Pattern performance heatmap
        plt.subplot(3, 2, 3)
        pattern_perf = df[df['Algorithm'] == 'Enhanced'].pivot_table(
            values='Time', index='Pattern', columns='Size', aggfunc='mean'
        )
        sns.heatmap(pattern_perf, annot=True, fmt='.3f', cmap='YlOrRd', 
                    cbar_kws={'label': 'Time (seconds)'})
        plt.title('Enhanced Sort Performance by Pattern')
        
        # 4. Relative performance heatmap
        plt.subplot(3, 2, 4)
        sizes = sorted(df['Size'].unique())
        patterns = df['Pattern'].unique()
        relative_perf = []
        
        for size in sizes:
            for pattern in patterns:
                subset = df[(df['Size'] == size) & (df['Pattern'] == pattern)]
                enhanced_time = subset[subset['Algorithm'] == 'Enhanced']['Time'].iloc[0]
                numpy_time = subset[subset['Algorithm'] == 'NumPy']['Time'].iloc[0]
                relative_perf.append({
                    'Size': size,
                    'Pattern': pattern,
                    'Relative': enhanced_time / numpy_time
                })
        
        rel_df = pd.DataFrame(relative_perf)
        rel_pivot = rel_df.pivot(index='Pattern', columns='Size', values='Relative')
        sns.heatmap(rel_pivot, annot=True, fmt='.2f', cmap='RdYlGn_r',
                    cbar_kws={'label': 'Time ratio (Enhanced/NumPy)'})
        plt.title('Performance Relative to NumPy')
        
        # 5. Time distribution
        plt.subplot(3, 2, 5)
        g = sns.violinplot(data=df, x='Pattern', y='Time', hue='Algorithm')
        plt.xticks(rotation=45)
        plt.title('Time Distribution by Pattern')
        plt.ylabel('Time (seconds)')
        g.legend(title='Algorithm', bbox_to_anchor=(1.05, 1), loc='upper left')
        
        # 6. Memory distribution
        plt.subplot(3, 2, 6)
        g = sns.violinplot(data=df, x='Pattern', y='Memory', hue='Algorithm')
        plt.xticks(rotation=45)
        plt.title('Memory Distribution by Pattern')
        plt.ylabel('Memory (MB)')
        g.legend(title='Algorithm', bbox_to_anchor=(1.05, 1), loc='upper left')
        
        # Adjust layout and save
        plt.tight_layout()
        plt.savefig('sorting_performance.png', dpi=300, bbox_inches='tight')
        plt.close()
        
        # Generate additional statistics
        stats_df = df.groupby(['Algorithm', 'Pattern']).agg({
            'Time': ['mean', 'std', 'min', 'max'],
            'Memory': ['mean', 'std']
        }).round(4)
        
        stats_df.to_csv('sorting_statistics.csv')

def main():
    visualizer = SortingVisualizer()
    sizes = [100, 1000, 10000]  # Reduced sizes for quicker testing
    patterns = ["random", "nearly_sorted", "reversed", "few_unique", "sorted", "pipeline"]
    
    print("Running comprehensive benchmarks...")
    results = visualizer.benchmark_sorting(sizes, patterns, repetitions=2)  # Reduced repetitions
    
    print("\nGenerating visualization...")
    visualizer.plot_performance_comparison(results)
    
    print("\nAnalysis complete! Check:")
    print("- 'sorting_performance.png' for detailed visualizations")
    print("- 'sorting_results.csv' for raw benchmark data")
    print("- 'sorting_statistics.csv' for detailed statistics")
    
    # Print summary of key findings
    print("\nKey Performance Findings:")
    print("-" * 60)
    
    # Calculate and display speedup over NumPy for different patterns
    for pattern in patterns:
        pattern_data = results[results['Pattern'] == pattern]
        enhanced = pattern_data[pattern_data['Algorithm'] == 'Enhanced']['Time'].mean()
        numpy = pattern_data[pattern_data['Algorithm'] == 'NumPy']['Time'].mean()
        speedup = numpy / enhanced
        print(f"{pattern:>15}: {speedup:>6.2f}x relative to NumPy")

if __name__ == "__main__":
    main() 